# pylint: skip-file
"""
.. _ref-oauth2-reference-server

==================================
OAuth 2.0 reference implementation
==================================
"""
import collections
import pathlib
from typing import Any

from ckms.core import Keychain

from .. import types
from ..authorizationserver import AuthorizationServer
from ..configfileclientrepository import ConfigFileClientRepository
from ..memorystorage import MemoryStorage
from ..staticsubjectepository import StaticSubjectRepository
from ..tokenissuer import TokenIssuer
from . import clients
from . import keys
from . import subject


__all__ = [
    'DEFAULT_ENCRYPTION_KEY',
    'DEFAULT_SIGNING_KEY',
    'alice',
    'bob',
    'clients',
    'keys',
    'keychain',
    'subject'
]

CLIENTS_DIR = pathlib.Path(__file__).parent.joinpath('etc/clients')

DEFAULT_ENCRYPTION_KEY = 'enc'

DEFAULT_SIGNING_KEY = 'sig'

alice = subject.Subject(
    sub="alice@example.unimatrixone.io",
    client_scope=collections.defaultdict(set, {
        'jwt': {"read", "write", 'openid', 'profile'}
    }),
    keys=[keys.alice]
)

bob = subject.Subject(
    sub="bob@example.unimatrixone.io",
    client_scope=collections.defaultdict(set, {
        'jwt': {"read", "write", "openid", "profile"},
        'confidential': {"read"}
    }),
    keys=[keys.bob]
)

trudy = subject.Subject(
    sub="trudy@example.unimatrixone.io",
    client_scope=collections.defaultdict(set, {
        'jwt': {"read", "write", "openid", "profile"},
        'confidential': {"read"}
    }),
    keys=[keys.trudy]
)

keychain: Keychain = Keychain()
keychain.configure({
    DEFAULT_ENCRYPTION_KEY: {
        'provider': 'local',
        'kty': "OKP",
        'alg': "EdDSA",
        'crv': 'Ed448',
        'tags': ['oauth2'],
        'use': 'enc',
    },
    DEFAULT_SIGNING_KEY: {
        'provider': 'local',
        'kty': "OKP",
        'alg': "EdDSA",
        'crv': 'Ed448',
        'tags': ['oauth2'],
        'use': 'sig',
    },
})

external_keychain: Keychain = Keychain()
external_keychain.configure({
    DEFAULT_ENCRYPTION_KEY: {
        'provider': 'local',
        'kty': "OKP",
        'crv': 'Ed448',
        'tags': ['oauth2'],
        'use': 'enc',
    },
    DEFAULT_SIGNING_KEY: {
        'provider': 'local',
        'kty': "OKP",
        'crv': 'Ed448',
        'tags': ['oauth2'],
        'use': 'sig',
    },
})

ClientRepository: type[types.IClientRepository] = ConfigFileClientRepository.new(
    base_path=CLIENTS_DIR
)

SubjectRepository: type[types.ISubjectRepository] = StaticSubjectRepository.new(
    model=subject.Subject,
    subjects={
        bob.sub: bob.dict(),
        alice.sub: alice.dict(),
        trudy.sub: trudy.dict()
    }
)


def get_oauth2_server(**kwargs: Any) -> AuthorizationServer:
    kwargs.setdefault('clients', ClientRepository)
    kwargs.setdefault('issuer', TokenIssuer)
    kwargs.setdefault('storage', MemoryStorage)
    kwargs.setdefault('subjects', SubjectRepository)
    return AuthorizationServer(**kwargs)


"""Declares :attr:`AuthorizationRequestClient`."""
from typing import Optional
from typing import TypeAlias

import fastapi
import pydantic

from .exceptions import Error
from .types import IClient
from .types import IClientRepository
from .params import ClientRepository


__all__ = [
    'AuthorizationRequestClient'
]


ClientIdentifierType: TypeAlias = pydantic.constr(max_length=63)


async def get_authorization_request_client(
    client_id: Optional[ClientIdentifierType] = fastapi.Query(
        default=None,
        title="Client ID",
        description="Identifies the client."
    ),
    clients: IClientRepository = ClientRepository
) -> IClient:
    if client_id is None or not await clients.exists(client_id):
        raise Error(
            error="invalid_request",
            error_description=(
                "The client_id parameter is required an must point to "
                "an existing client."
            )
        )
    return await clients.get(client_id)


AuthorizationRequestClient: IClient = fastapi.Depends(get_authorization_request_client)


"""Declares :class:`UpstreamProvider`."""
import fastapi

from .types import AuthorizationRequestParameters
from .types import IStorage
from .types import IUpstreamProvider


class UpstreamProvider(IUpstreamProvider):
    __module__: str = 'cbra.ext.oauth2'

    async def create_redirect(
        self,
        request: fastapi.Request,
        dto: AuthorizationRequestParameters
    ) -> str:
        """Create the redirect URI that initiates the authentication and/or
        authorization flow at the upstream identity provider.
        """
        raise NotImplementedError

    async def get_authorization_request(
        self,
        storage: IStorage,
        request: fastapi.Request
    ) -> AuthorizationRequestParameters:
        return await storage.get_authorization_request(
            request_id=self.get_authorization_request_id(request=request)
        )

    def get_client_redirect(
        self,
        request: fastapi.Request,
        dto: AuthorizationRequestParameters
    ) -> str:
        """Get the client redirect URL."""
        return dto.authorize()

    async def on_return(
        self,
        storage: IStorage,
        request: fastapi.Request
    ) -> fastapi.responses.RedirectResponse:
        params = await self.get_authorization_request(
            storage=storage,
            request=request
        )
        await self.process_response(request, params)
        return fastapi.responses.RedirectResponse(
            url=self.get_client_redirect(
                request=request,
                dto=params
            ),
            status_code=303
        )

    async def process_response(
        self,
        request: fastapi.Request,
        params: AuthorizationRequestParameters
    ) -> None:
        """Handles the response data from the upstream provider."""
        raise NotImplementedError

    def reverse(self, request: fastapi.Request) -> str:
        """Get the return URL to which the upstream identity
        provider must redirect.
        """
        return request.url_for(f'oauth2.callback', provider=self.name)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
from typing import Any

from .basemessagetest import BaseMessageTest


class TestProductUpdated(BaseMessageTest):
    event_type: str = 'updated'
    resource_type: str = 'product'

    def get_request_params(self) -> dict[str, Any]:
        return {
            'headers': {
                'content-type': "application/json",
                'x-wc-webhook-delivery-id': "59edde97ec57fb8c05832b525fb27e9a",
                'x-wc-webhook-event': "updated",
                'x-wc-webhook-id': "57",
                'x-wc-webhook-resource': "product",
                'x-wc-webhook-signature': "oQ+wJoz5g/S9HJEwBIzVQYT1SQvlXJ3khoxeJbQroYo=",
                'x-wc-webhook-source': "https://test-wordpress.molano.nl/",
                'x-wc-webhook-topic': "product.updated"
            },
            'content': (
                b'{"id":57,"name":"variants 4 - foo, red","slug":"variants'
                b'-lijp-red-foo","permalink":"https:\/\/test-wordpress.mol'
                b'ano.nl\/product\/variants\/?attribute_teset=foo&attribut'
                b'e_color=red","date_created":"2022-05-03T13:39:45","date_'
                b'created_gmt":"2022-05-03T13:39:45","date_modified":"2022'
                b'-06-03T13:15:43","date_modified_gmt":"2022-06-03T13:15:4'
                b'3","type":"variation","status":"publish","featured":fals'
                b'e,"catalog_visibility":"visible","description":"<p>Tst<\/'
                b'p>\\n","short_description":"","sku":"XXXBAZ","price":"",'
                b'"regular_price":"","sale_price":"","date_on_sale_from":n'
                b'ull,"date_on_sale_from_gmt":null,"date_on_sale_to":null,'
                b'"date_on_sale_to_gmt":null,"on_sale":false,"purchasable"'
                b':false,"total_sales":"0","virtual":false,"downloadable":'
                b'false,"downloads":[],"download_limit":-1,"download_expir'
                b'y":-1,"external_url":"","button_text":"","tax_status":"t'
                b'axable","tax_class":"","manage_stock":true,"stock_quanti'
                b'ty":20,"backorders":"no","backorders_allowed":false,"bac'
                b'kordered":false,"low_stock_amount":null,"sold_individual'
                b'ly":false,"weight":"","dimensions":{"length":"","width":'
                b'"","height":""},"shipping_required":true,"shipping_taxab'
                b'le":true,"shipping_class":"","shipping_class_id":0,"revi'
                b'ews_allowed":false,"average_rating":"0.00","rating_count'
                b'":0,"upsell_ids":[],"cross_sell_ids":[],"parent_id":46,"'
                b'purchase_note":"","categories":[],"tags":[],"images":[],'
                b'"attributes":[{"id":0,"name":"teset","option":"foo"},{"i'
                b'd":0,"name":"color","option":"red"}],"default_attributes'
                b'":[],"variations":[],"grouped_products":[],"menu_order":'
                b'1,"price_html":"","related_ids":[],"meta_data":[],"stock'
                b'_status":"instock","_links":{"self":[{"href":"https:\/\/'
                b'test-wordpress.molano.nl\/wp-json\/wc\/v3\/products\/57"'
                b'}],"collection":[{"href":"https:\/\/test-wordpress.molan'
                b'o.nl\/wp-json\/wc\/v3\/products"}],"up":[{"href":"https:'
                b'\/\/test-wordpress.molano.nl\/wp-json\/wc\/v3\/products\/'
                b'46"}]}}'
            )
        }


"""Declares :class:`YAMLRenderer`."""
import typing

import pydantic
import yaml

from .json import JSONRenderer


class YAMLRenderer(JSONRenderer):
    """
    Renderer which serializes to YAML.
    """
    __module__: str = 'cbra.renderers'
    format: str = 'yaml'
    media_type: str = 'application/yaml'
    response_media_type: str = "application/yaml"
    #encoder_class = encoders.JSONEncoder

    def has_content(self) -> bool:
        return True

    def render(
        self,
        data: dict[str, typing.Any] | pydantic.BaseModel,
        renderer_context: dict[str, typing.Any] | None = None
    ) -> bytes:
        """Render `data` into YAML, returning a bytestring."""
        if data is None:
            return b''
        assert data is not None # nosec
        serialized = yaml.safe_dump( # type: ignore
            data,
            indent=self.get_indent(self.accepted, renderer_context or {}),
            default_flow_style=False
        )
        return '---\n' + serialized # type: ignore



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeAlias

from .deletedevent import DeleteEvent
from .event import Event
from .ping import Ping
from .orderevent import OrderEvent
from .productevent import ProductEvent


__all__: list[str] = [
    'DeleteEvent',
    'Event',
    'EventType',
    'OrderEvent',
    'ProductEvent',
    'Ping'
]

EventType: TypeAlias = (
    OrderEvent | ProductEvent | DeleteEvent | Ping
)


# type: ignore
"""Declares :class:`ServerTest`."""
import typing
import urllib.parse

import cbra
import fastapi
import pytest
import pytest_asyncio
import unimatrix.lib.http
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken
from httpx import AsyncClient
from httpx import Response

from cbra.ext.jwks import JWKSEndpoint
from cbra.ext.oauth2 import AuthorizationEndpoint
from cbra.ext.oauth2 import ConfigFileClientRepository
from cbra.ext.oauth2 import MemoryClientRepository
from cbra.ext.oauth2 import MemoryStorage
from cbra.ext.oauth2 import MemorySubjectRepository
from cbra.ext.oauth2 import AuthorizationServer
from cbra.ext.oauth2 import RFC9068Principal
from cbra.ext.oauth2 import TokenRequestHandler
from cbra.ext.oauth2 import TokenIssuer
from cbra.ext.oauth2 import ref
from cbra.ext.oauth2.types import IClient
from cbra.ext.oauth2.types import IClientRepository
from cbra.ext.oauth2.types import ISubject
from cbra.ext.oauth2.types import ISubjectRepository
from cbra.types import IPrincipal


class ServerTest:
    base_url: str = "https://cbra-oauth2.localhost"
    authorize_url: str = f'{base_url}/oauth2/authorize'
    authorize_status_code: int = 303
    token_url: str = f'{base_url}/oauth2/token'
    redirect_uri: str = f"{base_url}/callback"
    jar_url: str = f"{base_url}/oauth2/par"
    message: str = "Hello world!"
    client_id: str = None

    @pytest.fixture
    def handlers(self) -> typing.List[typing.Any]:
        return []

    @pytest_asyncio.fixture
    async def app(
        self,
        server: AuthorizationServer,
        handlers: typing.List[typing.Any],
        keychain: Keychain
    ) -> cbra.Application:
        await keychain
        app = cbra.Application(
            keychain=keychain
        )
        app.add_api_route('/rfc9068', endpoint=self.handle_rfc9068_token)
        app.add(server, base_path='/oauth2')
        for handler_class, base_path in handlers:
            app.add(handler_class, base_path=base_path)
        await app.on_startup()
        await server.setup_authorization_server()
        return app

    def assert_error(self, response: Response, error: str, status_code: int = 400):
        assert response.status_code == status_code, response.status_code
        assert response.json()['error'] == error, response.json()

    def assert_redirects(self, response: Response, uri: str):
        u1 = urllib.parse.urlparse(response.headers['Location'])
        u2 = urllib.parse.urlparse(uri)
        assert u1[:3] == u2[:3]

    def assert_query_parameter(self, url: str, name: str, value: str):
        p = urllib.parse.urlparse(url)
        q = unimatrix.lib.http.parse_qs(p.query)
        assert name in q, f'{name} not in {repr(q)}'
        assert q[name] == value

    async def get_current_client(self) -> IClient:
        return ConfigFileClientRepository().get(self.client_id)

    async def get_client_keychain(self, key: str) -> Keychain:
        keychain = Keychain()
        keychain.configure({
            'sig': {
                'provider': "local",
                'kty': "OKP",
                'alg': "EdDSA",
                'key': {'path': f"pki/{key}.key"},
                'crv': "Ed448",
                'use': 'sig'
            }
        })
        await keychain
        return keychain

    async def get_client_codec(
        self,
        keychain: Keychain
    ) -> PayloadCodec:
        return PayloadCodec(
            signer=keychain
        )

    async def handle_rfc9068_token(
        self,
        principal: IPrincipal = fastapi.Depends(RFC9068Principal())
    ):
        assert principal.has_scope("read")
        return {"message": self.message}

    def assert_rfc9068(self, token: str, claim: str, value: typing.Any):
        header, payload, signature = str.split(token, '.')
        jwt = JSONWebToken.frompayload(payload)
        claims = jwt.dict()
        assert claim in claims
        assert claims[claim] == value



# Copyright (C) 2022 Cochise Ruhulessin <cochiseruhulessin@gmail.com>
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
"""Declares :class:`IPrincipal`."""
import typing

import cbra.types


class IPrincipal(cbra.types.IPrincipal):
    """A :class:`~cbra.types.IPrincipal` interface with additional
    methods.
    """
    __module__: str = 'cbra.ext.oauth2.types'

    #: The OAuth 2.0 client that identified this principal.
    client_id: str

    #: The scope that was granted to the principal.
    scope: typing.Set[str]

    def is_authenticated(self) -> bool:
        return True


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
from typing import Any

from .basemessagetest import BaseMessageTest


class TestOrderCreated(BaseMessageTest):
    event_type: str = 'created'
    resource_type: str = 'order'

    def get_request_params(self) -> dict[str, Any]:
        return {
            'headers': {
                'content-type': "application/json",
                'x-wc-webhook-delivery-id': "49b041fc9aee3f9aca0c0b5173c2b0d7",
                'x-wc-webhook-event': "created",
                'x-wc-webhook-id': "61",
                'x-wc-webhook-resource': "order",
                'x-wc-webhook-signature': "iuU91bAwFYjqebtl3dsJhJFR0Dst2VK2Yldz+Kf9VYk=",
                'x-wc-webhook-source': "https://test-wordpress.molano.nl/",
                'x-wc-webhook-topic': "order.created",
            },
            'content': (
                b'{"id":76,"parent_id":0,"status":"pending","currency":"EUR","version":"'
                b'5.9.1","prices_include_tax":false,"date_created":"2022-06-03T20:28:09"'
                b',"date_modified":"2022-06-03T20:28:58","discount_total":"0.00","discou'
                b'nt_tax":"0.00","shipping_total":"0.00","shipping_tax":"0.00","cart_tax'
                b'":"0.00","total":"0.00","total_tax":"0.00","customer_id":1,"order_key"'
                b':"wc_order_tpYsgYNGJe8OD","billing":{"first_name":"","last_name":"","c'
                b'ompany":"","address_1":"","address_2":"","city":"","state":"","postcod'
                b'e":"","country":"","email":"cochise.ruhulessin@unimatrixone.io","phone'
                b'":""},"shipping":{"first_name":"","last_name":"","company":"","address'
                b'_1":"","address_2":"","city":"","state":"","postcode":"","country":"",'
                b'"phone":""},"payment_method":"","payment_method_title":"","transaction'
                b'_id":"","customer_ip_address":"","customer_user_agent":"","created_via'
                b'":"admin","customer_note":"","date_completed":null,"date_paid":null,"c'
                b'art_hash":"","number":"76","meta_data":[],"line_items":[{"id":4,"name"'
                b':"Test Product p","product_id":39,"variation_id":0,"quantity":1,"tax_c'
                b'lass":"","subtotal":"0.00","subtotal_tax":"0.00","total":"0.00","total'
                b'_tax":"0.00","taxes":[],"meta_data":[],"sku":"XXXTEST","price":0,"pare'
                b'nt_name":null}],"tax_lines":[],"shipping_lines":[],"fee_lines":[],"cou'
                b'pon_lines":[],"refunds":[],"date_created_gmt":"2022-06-03T20:28:09","d'
                b'ate_modified_gmt":"2022-06-03T20:28:58","date_completed_gmt":null,"dat'
                b'e_paid_gmt":null,"currency_symbol":"\\u20ac","_links":{"self":[{"href":'
                b'"https:\\/\\/test-wordpress.molano.nl\\/wp-json\\/wc\\/v3\\/orders\\/76"}],"c'
                b'ollection":[{"href":"https:\\/\\/test-wordpress.molano.nl\\/wp-json\\/wc\\/'
                b'v3\\/orders"}],"customer":[{"href":"https:\\/\\/test-wordpress.molano.nl'
                b'\\/wp-json\\/wc\\/v3\\/customers\\/1"}]}}'
            )
        }


from ckms.core import Keychain

import cbra
from cbra.ext.jwks import JWKSEndpoint


__all__ = ['get_asgi_application']


def get_asgi_application() -> cbra.Application:
    keychain = Keychain()
    keychain.configure({ # type: ignore
        'sig': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'RS256',
            'key': {'length': 2048},
            'tags': ['server'],
        },
        'enc': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'RSA-OAEP-256',
            'key': {'length': 2048},
            'tags': ['server'],
        }
    })
    app = cbra.Application(
        keychain=keychain
    )
    app.add(JWKSEndpoint)
    return app


app = get_asgi_application()
if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)



# Copyright (C) 2022 Cochise Ruhulessin <cochiseruhulessin@gmail.com>
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
"""Declares :class:`IPrincipal`."""
from ckms.types import ClaimSet


class IPrincipal:
    """Specifies the interface for all principal implementations."""

    #: The subject identifier; establishes the identity of the current user,
    #: application or service account that issued a request to the HTTP
    #: server.
    sub: int | str | None

    @classmethod
    def fromclaimset(
        cls,
        claims: ClaimSet
    ) -> 'IPrincipal':
        """Instantiate a new :class:`IPrincipal` using the claims provided with a
        JSON Web Token (JWT).
        """
        raise NotImplementedError

    def is_authenticated(self) -> bool:
        """Return a boolean indicating if the principal is authenticated. Most
        commonly this will imply that :attr:`sub` is not ``None``.
        """
        raise NotImplementedError

    def has_scope(
        self,
        scope: str | set[str]
    ) -> bool:
        """Return a boolean indicating if the :class:`IPrincipal` is
        authorized to use the given `scope`.
        """
        raise NotImplementedError

    def __bool__(self) -> bool:
        return self.is_authenticated()


"""Declares constants used by the :mod:`cbra.session` module."""
from cbra.conf import settings # type: ignore


SESSION_COOKIE_AUDIENCE: str = (
    getattr(settings, 'SESSION_COOKIE_AUDIENCE', None) # type: ignore
)


SESSION_COOKIE_NAME: str = (
    getattr(settings, 'SESSION_COOKIE_NAME', None) # type: ignore
    or 'session-claims'
)

SESSION_SIGNING_KEYS: str = (
    getattr(settings, 'SESSION_SIGNING_KEY', None) # type: ignore
    or ['session-signer-1', 'session-signer-2']
)

SESSION_ENCRYPTION_KEYS: str = (
    getattr(settings, 'SESSION_ENCRYPTION_KEY', None) # type: ignore
    or ['session-encrypter-1']
)

SESSION_ENCRYPTION_ALG: str = (
    getattr(settings, 'SESSION_ENCRYPTION_ALG', None) # type: ignore
    or 'A256GCM'
)



"""Declares common OAuth 2.0 exceptions."""
import functools
import typing

import fastapi
import fastapi.responses
import pydantic
from unimatrix.exceptions import CanonicalException

from cbra.exceptions import BaseRedirectingException


__all__ = [
    'ClientAuthenticationFailed',
    'Error',
    'ErrorModel',
    'InvalidClient',
    'LoginRequired',
    'MissingRedirectURL',
    'RedirectForbidden',
    'SubjectDoesNotExist',
    'TokenExchangeTargetRejected',
    'UnsupportedTokenTypeRequested',
    'UnsupportedResponseType',
    'UnsupportedSubjectTokenType',
]


class LoginRequired(BaseRedirectingException):
    """Raised when the resource owner needs to authenticate with
    the authorization server.
    """
    __module__: str = 'cbra.ext.oauth2.exceptions'


class ErrorModel(pydantic.BaseModel):
    error: str
    error_description: typing.Optional[str] = None
    error_uri: typing.Optional[str] = None


class Error(Exception):
    error: str = "invalid_request"
    error_description: str
    mode: str = 'client'
    state: str | None = None

    @classmethod
    def as_redirect(cls, redirect_uri: typing.Any) -> 'Error':
        return cls(redirect_uri=redirect_uri)

    @classmethod
    def catch(
        cls,
        error: typing.Type[BaseException],
        description: typing.Optional[str] = None
    ) -> typing.Callable[..., typing.Any]:
        def decorator_factory(func: typing.Any) -> typing.Callable[..., typing.Any]:
            @functools.wraps(func)
            async def decorated(*args: typing.Any, **kwargs: typing.Any):
                try:
                    return await func(*args, **kwargs)
                except error as exc:
                    msg = description
                    if isinstance(exc, CanonicalException) and not msg:
                        msg = exc.message or exc.code
                    if callable(description):
                        msg = typing.cast(str, description(exc))
                    raise cls(error_description=msg)
            return decorated
        return decorator_factory

    def __init__(self, **kwargs: str | typing.Any):
        kwargs.setdefault('error', self.error)
        kwargs.setdefault('error_description', self.get_error_description(**kwargs))
        self.mode = kwargs.pop('mode', self.mode)
        self.state = kwargs.pop('state', None)
        self._redirect_uri = kwargs.pop('redirect_uri', None)
        self._error = ErrorModel(
            error=typing.cast(str, kwargs['error']),
            error_description=typing.cast(str, kwargs['error_description'])
        )

    async def as_response(
        self,
        request: fastapi.Request,
        mode: str | None = None,
        state: str | None = None
    ) -> fastapi.responses.JSONResponse | fastapi.responses.RedirectResponse:
        """Return a :class:`fastapi.responses.Response` instance
        representing the error.
        """
        state = state or self.state
        mode = mode or self.mode
        if mode == 'client' or not self._redirect_uri:
            return fastapi.responses.JSONResponse(
                self._error.dict(exclude_defaults=True),
                status_code=400
            )
        else:
            params = {
                'error': self._error.error,
                'error_description': self._error.error_description
            }
            if state is not None:
                params['state'] = state
            return fastapi.responses.RedirectResponse(
                self._redirect_uri.authorize(**params),
                status_code=303
            )

    def get_error_description(self, **kwargs: str) -> str:
        """Return the error description."""
        return kwargs.get('error_description') or self.error_description

    def must_redirect(self) -> bool:
        return self.mode == 'redirect'


class ClientAuthenticationFailed(Error):
    """Raised when the client credentials could not be authenticated."""
    __module__: str = 'cbra.ext.oauth2'
    error: str = "invalid_request"
    error_description: str = (
        "The client could not be authenticated using the given credentials."
    )


class InvalidGrant(Error):
    __module__: str = 'cbra.ext.oauth2'
    error: str = "invalid_grant"


class AssertionReplayed(InvalidGrant):
    """Raised when the credential is not valid because its replayed."""
    __module__: str = 'cbra.ext.oauth2'
    error_description: str = (
        "The assertion provided by the client was used previously to "
        "obtain an access token. Do not reuse assertions."
    )


class RedirectForbidden(Error):
    """Raised when the ``redirect_uri`` parameter is not valid
    for the given client.
    """
    error: str = "invalid_request"
    error_description: str = (
        "The client does not allow redirects to the URL supplied as "
        "the redirect_uri query parameter."
    )


class MissingRedirectURL(Error):
    error: str = "invalid_request"
    error_description: str = (
        "The `redirect_uri` parameter was not provided and the client "
        "did not specify a default."
    )


class RequestNotSupported(Error):
    error: str = "request_not_supported"
    error_description: str = (
        "The server does not support the use of the `request` or "
        "`request_uri` parameters."
    )


class InvalidClient(Error):
    error: str = "invalid_request"
    error_description: str = (
        "Specify the `client_id` parameter in the request body "
        "if the client is public, else provide credentials with "
        "the `Authorization` header."
    )


class InvalidScope(Error):
    error: str = "invalid_scope"
    error_description: str = (
        "The requested scope exceeds the scope that was granted by the "
        "resource owner."
    )


class ForbiddenScope(Error):
    error: str = "invalid_scope"
    error_description: str = (
        "The client application is not allowed or not allowing use of "
        "the requested scope."
    )


class ClientDoesNotExist(InvalidClient):
    """Like :exc:`InvalidClient`, but explicitely indicates that the
    client does not exist.
    """
    __module__: str = 'cbra.ext.oauth2'
    error_description: str = "The specified client does not exist."


class ClientAuthenticationRequired(InvalidClient):
    """Raised when there is an issued with authenticating a condfidential
    OAuth 2.0 client."""
    __module__: str = 'cbra.ext.oauth2'
    error: str = "invalid_client"
    error_description: str = "The specified client requires authentication."


class InvalidClientAuthenticationScheme(ClientAuthenticationRequired):
    """Raised when the client used the wrong authentication scheme."""
    __module__: str = 'cbra.ext.oauth2'

    def __init__(self, using: str, required: str):
        super().__init__(
            error_description=(
                "Invalid authentication scheme. The server supports "
                f"{required} authentication for confidential clients."
            )
        )


class NoSuchClient(InvalidClient):
    """Raised during a request to the Token Endpoint when the client
    could not be determined from any factor.
    """
    __module__: str = 'cbra.ext.oauth2'
    error: str = "invalid_client"
    error_description: str = (
        "The client specified by the request parameters does not exist."
    )


class UnsupportedTokenTypeRequested(Error):
    """Raised when an unsupported token type is requested."""

    def __init__(self, requested: str, supported: list, **kwargs):
        Error.__init__(
            self,
            error="invalid_request",
            error_description=(
                "The server is unable to issue a token of the requested type: "
                f"{requested}. Supported token types are: "
                f"{str.join(',', sorted(supported))}."
            ),
            **kwargs
        )


class UnsupportedSubjectTokenType(Error):
    """Raised when an unsupported token type is provided."""

    def __init__(self, given: str, supported: list, **kwargs):
        Error.__init__(
            self,
            error="invalid_request",
            error_description=(
                "The server is unable to process a token of the provided type: "
                f"{given}. Supported token types are: "
                f"{str.join(',', sorted(supported))}."
            ),
            **kwargs
        )


class SubjectDoesNotExist(Error):
    """Raises when a lookup for a :term:`Subject` is attempted
    but the search query did not produce a result.
    """
    __module__: str = 'cbra.ext.oauth2'
    error: str = "invalid_request"
    error_description: str = (
        "The principal attached to the request did not resolve "
        "to a subject."
    )


class TokenExchangeTargetRejected(Error):
    """Raised when the server does not allow a specific audience or resource that
    is requested during a token exchange.
    """

    def __init__(self, target: str, message: typing.Optional[str] = None, **kwargs):
        Error.__init__(
            self,
            error="invalid_target",
            error_description=message or (
                f"The target {target} rejected the claims, scope or subject of "
                "the token exchange."
            ),
            **kwargs
        )


class UnsupportedResponseType(Error):
    __module__: str = 'cbra.ext.oauth2'
    error: str = (
        "The requested response type is not supported by the "
        "client or server."
    )
    error_description: str = (
        "Consult the documentation for the supported "
        "response types."
    )
    mode: str = 'redirect'


class CrossOriginNotAllowed(Error):
    error: str = "invalid_origin"
    error_description: str = (
        "Security measures prevent the authorization server from "
        "responding to requests from this domain, because it has "
        "not been whitelisted by the system administrator for use "
        "with this application."
    )
    mode: str = 'redirect'


# pylint: skip-file
from typing import AsyncGenerator

import fastapi
import pytest
import pytest_asyncio
from ckms.types import JSONWebToken
from ckms.types import MalformedPayload
from httpx import AsyncClient

from cbra import Application
from cbra.session import CookieSession
from cbra.session.const import SESSION_COOKIE_NAME


async def get(session: CookieSession = fastapi.Depends()):
    """Returns the claims that are currently in the session."""
    await session # type: ignore
    return session.claims


@pytest.fixture
def server(app: Application) -> Application:
    app.add_api_route(path='/get', endpoint=get)
    return app


@pytest_asyncio.fixture # type: ignore
async def client(server: Application) -> AsyncGenerator[AsyncClient, None]:
    params = {
        'app': server,
        'base_url': "https://cbra.localhost"
    }
    async with AsyncClient(**params) as client:
        yield client


@pytest_asyncio.fixture # type: ignore
async def session_cookie(client: AsyncClient) -> str:
    response = await client.get('/get') # type: ignore
    return response.cookies[SESSION_COOKIE_NAME]


@pytest.mark.asyncio
async def test_session_is_created(client: AsyncClient):
    response = await client.get('/get') # type: ignore
    claims = response.json()
    assert SESSION_COOKIE_NAME in response.cookies
    assert 'iat' in claims
    assert len(claims) == 1


@pytest.mark.asyncio
async def test_session_is_created_on_malformed(client: AsyncClient):
    response = await client.get( # type: ignore
        url='/get',
        cookies={
            SESSION_COOKIE_NAME: 'foo'
        }
    )
    claims = response.json()
    assert SESSION_COOKIE_NAME in response.cookies
    assert 'iat' in claims
    assert len(claims) == 1


@pytest.mark.asyncio
async def test_session_is_persisted(client: AsyncClient):
    response = await client.get('/get') # type: ignore
    claims = response.json()
    assert SESSION_COOKIE_NAME in response.cookies
    assert 'iat' in claims
    assert len(claims) == 1
    iat = claims['iat']
    response = await client.get('/get') # type: ignore
    claims = response.json()
    assert iat == claims.get('iat')


@pytest.mark.asyncio
async def test_server_can_decode_session(
    server: Application,
    session_cookie: str
):
    _, jwt = await server.codec.jwt(
        session_cookie.encode(),
        accept="jwt+session"
    )
    claims = jwt.dict()
    assert isinstance(jwt, JSONWebToken)
    assert 'iat' in claims
    assert len(claims) == 1


@pytest.mark.asyncio
async def test_server_requires_typed_jwt(
    server: Application,
    session_cookie: str
):
    with pytest.raises(MalformedPayload):
        await server.codec.jwt(
            session_cookie.encode(),
            accept="jwt"
        )


# pylint: skip-file
import pytest

from ..dependency import Dependency


@pytest.mark.asyncio
async def test_polymorphic_symbol(symbol, package_name):
    spec = Dependency(
        spec={
            'type': 'symbol',
            'name': "SymbolDependency",
            'qualname': f"{package_name}.conftest.SYMBOL"
        }
    )
    assert await spec.resolve() == symbol


@pytest.mark.asyncio
async def test_polymorphic_file():
    spec = Dependency(
        spec={
            'type': 'file',
            'name': 'ExampleDependency',
            'path': 'VERSION'
        }
    )
    with open('VERSION', 'rb') as f:
        assert await spec.resolve() == f.read()



# type: ignore
"""Declares :class:`ClientConfig`."""
import secrets
import typing

from ckms.core.models import JSONWebSignature
from ckms.types import IKeychain
from ckms.types import JSONWebKey
from ckms.types import JSONWebKeySet
from ckms.types import JSONWebToken
from ckms.types import MalformedPayload
from ckms.jose import PayloadCodec

from cbra.utils import current_timestamp
from .baseclient import BaseClient
from .clientsubject import ClientSubject
from .exceptions import Error
from .exceptions import InvalidGrant
from .exceptions import MissingRedirectURL
from .exceptions import RedirectForbidden
from .types import BaseGrant
from .types import ISubject
from .types import GrantType
from .types import RedirectURL
from .types import ResponseType


class ClientConfig(BaseClient):
    InvalidCredential: type[Exception] = type('InvalidCredential', (Exception,), {})

    allow_key_discovery: bool = False

    allow_multiple_audiences: bool = False

    allowed_audience: typing.Set[str] = set()

    allowed_origins: typing.Set[str] = set()

    client_id: str

    client_secret: str | None = None

    default_redirect_url: RedirectURL | None = None

    grant_types_supported: list[GrantType] = []

    jwks_uri: str | None = None

    keys: list[JSONWebKey] = []

    redirect_uris: list[RedirectURL] = []

    response_types: set[ResponseType] = set()

    scope: set[str] = set()

    @property
    def jwks(self) -> JSONWebKeySet:
        return JSONWebKeySet(keys=self.keys)

    def __init__(self, **kwargs: typing.Any):
        redirect_uris = kwargs.setdefault('redirect_uris', [])
        if 'default_redirect_url' in kwargs:
            redirect_uris.append(kwargs['default_redirect_url'])
        super().__init__(**kwargs)

    def allows_audience(self, issuer: str, audience: typing.Set[str]) -> bool:
        """Return a boolean indicating if the client allows the issueing
        of access token for the given audiences set `audience`.
        """
        return (audience <= (self.allowed_audience | {issuer}))

    def allows_grant_type(self, grant_type: GrantType) -> bool:
        """Return a boolean indicating if the client allows the
        given grant type.
        """
        return grant_type.value in self.grant_types_supported

    def allows_origin(self, origin: str | None) -> bool:
        """Return a boolean indicating if the client allows the given origin in the
        context of Cross-Origin Resource Sharing (CORS).
        """
        return origin is None or origin in self.allowed_origins

    def allows_response_type(self, response_type: ResponseType) -> bool:
        """Return a boolean indicating if the client allows the given response
        type.
        """
        return response_type in self.response_types

    def allows_scope(self, scope: set[str]) -> bool:
        """Return a boolean indicating if the client allows the given scope."""
        return bool(set(scope) <= self.scope)

    def as_subject(self) -> ISubject:
        """Return a :class:`~cbra.ext.oauth2.types.ISubject` implementation
        representing the client itself. For use with the client credentials
        grant.
        """
        return ClientSubject(client_id=self.client_id)

    def can_issue_multiple(self) -> bool:
        """Return a boolean indicating if the client allows issueing
        access tokens to multiple audiences.
        """
        return self.allow_multiple_audiences

    def can_redirect(self, url: typing.Union[str, RedirectURL]) -> bool:
        """Return a boolean indicating if the client allows
        redirects to the given `url`.
        """
        return url in self.redirect_uris

    def get_default_redirect(self) -> str:
        """Return the default redirect URL, if specified."""
        return self.default_redirect_url

    def get_redirect_url(
        self,
        url: str | RedirectURL | None = None,
        fatal: bool = True
    ) -> RedirectURL | None:
        """Return a :class:`RedirectURL` instance that the server is
        allowed to redirect to.
        """
        # If multiple redirect URIs have been registered, the client
        # MUST include a redirect URI with the authorization request
        # using the "redirect_uri" request parameter (OAuth 2.1 draft,
        # 2.3.3).
        try:
            if url is None and len(self.redirect_uris) > 1:
                raise MissingRedirectURL
            if url is None and self.redirect_uris:
                url = self.redirect_uris[0]
            if not self.redirect_uris:
                raise RedirectForbidden

            if not isinstance(url, RedirectURL):
                url = RedirectURL.validate(url)
            if not self.can_redirect(url):
                raise RedirectForbidden
        except Error as exception:
            if fatal:
                raise
            url = None
        return url

    async def issue_token(
        self,
        codec: PayloadCodec,
        algorithm: str,
        using: str,
        issuer: str,
        audience: typing.Union[str, typing.Set[str]],
        subject: ISubject,
        ttl: int,
        scope: typing.Set[str]
    ) -> str:
        """Issue an access token encoded as a JSON Web Token (JWT)
        as defined in :rfc:`9068`. Return a string containing the
        token.
        """
        claims = {
            'aud': list(audience),
            'client_id': self.client_id,
            'iss': issuer,
            'jti': secrets.token_urlsafe(24),
            'scope': str.join(' ', sorted(scope)),
            'sub': subject.sub,
        }

        # The scope claim can be removed if it is empty because "if an
        # authorization request includes a scope parameter, the
        # corresponding issued JWT access token SHOULD include a "scope"
        # claim as defined in Section 4.2 of [RFC8693]." (RFC 9068, 2.2.3)
        if not scope:
            claims.pop('scope')

        now = claims['iat'] = current_timestamp()
        claims.update({
            'exp': now + ttl,
            'nbf': now,
        })
        return await codec.encode(
            JSONWebToken(**claims),
            content_type="at+jwt",
            signers=[using]
        )

    def is_confidential(self) -> bool:
        """Return a boolean indicating if the client is confidential."""
        return bool(self.keys or self.client_secret)

    def is_public(self) -> bool:
        """Return a boolean indicating if the client is public."""
        return self.client_secret is None and not self.keys

    async def verify_jws(
        self,
        jws: JSONWebSignature
    ) -> bool:
        """Verify a :class:`~ckms.jose.JSONWebSignature` using the public
        keys associated to the client.
        """
        return await jws.verify(self.jwks)

    async def verify_jwt(
        self,
        token: typing.Union[bytes, str],
        decrypter: typing.Optional[IKeychain],
        compact_only: bool = False,
        audience: str | None = None
    ) -> JSONWebToken:
        """Verifies a JSON Web Token (JWT) using the public keys of the
        client.

        Args:
            token (byte-sequence or string): the JWT to deserialize and
                verify.
            decrypter: a :class:`~IKeychain` implementation that has the
                private keys to decrypt, if the payload is a JWE.
            compact_only (bool): indicates if only compact deserialization
                is allowed. If `compact_only` is ``True`` and the token
                is not compact serialized, an exception is raised.

        Returns
            A :class:`dict` containing the claims in the JWT.
        """
        codec = PayloadCodec(decrypter=decrypter, verifier=self.jwks)
        if isinstance(token, str): # pragma: no cover
            token = str.encode(token, 'ascii')
        jwt = await codec.decode(token, accept="jwt")
        if not isinstance(jwt, JSONWebToken):
            raise MalformedPayload(message="A JSON Web Token (JWT) is expected.")
        jwt.verify(audience={audience} if audience else set())
        return jwt



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
from typing import Any

from .basemessagetest import BaseMessageTest


class TestOrderDeleted(BaseMessageTest):
    event_type: str = 'deleted'
    resource_type: str = 'order'

    def get_request_params(self) -> dict[str, Any]:
        return {
            'headers': {
                'content-type': "application/json",
                'x-wc-webhook-delivery-id': "b81aaee861e1ac036579255f2b61e495",
                'x-wc-webhook-event': "deleted",
                'x-wc-webhook-id': "63",
                'x-wc-webhook-resource': "order",
                'x-wc-webhook-signature': "FoPlvaJLF/h0tKiA6Xo6rw14pWwLQ84c+xrbByDKY7M=",
                'x-wc-webhook-source': "https://test-wordpress.molano.nl/",
                'x-wc-webhook-topic': "order.deleted",
            },
            'content': b'{"id":76}'
        }


# pylint: skip-file
import inspect
import typing

from aorta import Command
from aorta import Event
from fastapi import Body
from fastapi import Depends
from fastapi import Header
from fastapi import Request
from fastapi import Path
from fastapi import Query
from fastapi.concurrency import AsyncExitStack
from fastapi.responses import Response
import uvicorn

from .application import Application
from .dependantrunner import DependantRunner
from .dependencyreceiver import DependencyReceiver
from .endpoint import Endpoint
from .endpointmetaclass import EndpointMetaclass
from .injectedmessagehandler import CommandHandler
from .injectedmessagehandler import EventListener
from .messagepublisher import _default as MessagePublisher # type: ignore
from .options import Options
from .resource import Resource
from .useragent import UserAgent
from . import cors
from . import renderers
from . import transaction
from . import types
from . import utils
from . import query


__all__ = [
    'cors',
    'renderers',
    'run',
    'transaction',
    'query',
    'types',
    'utils',
    'Application',
    'Body',
    'Command',
    'CommandHandler',
    'DependantRunner',
    'DependencyReceiver',
    'Endpoint',
    'EndpointMetaclass',
    'Event',
    'EventListener',
    'Header',
    'MessagePublisher',
    'Options',
    'Path',
    'Query',
    'Resource',
    'Response',
    'UserAgent',
]


def run(
    app: typing.Union[str, Application],
    *args: typing.Any,
    **kwargs: typing.Any
):
    """Like :func:`uvicorn.run`, but with additional integrations with
    :mod:`cbra`.
    """
    from cbra.conf import settings
    kwargs.setdefault('ssl_certfile', settings.LOCALHOST_SSL_CRT)
    kwargs.setdefault('ssl_keyfile', settings.LOCALHOST_SSL_KEY)
    return uvicorn.run(app, *args, **kwargs)


async def run_dependant(
    app: Application,
    dependant: typing.Callable,
    scope: dict = {},
    path: str = None
):
    """Run callable `dependant` and ensure that it is invoked with all its
    dependencies resolved.
    """
    runner = DependantRunner()
    async with AsyncExitStack() as stack:
        request = Request(
            scope={
                'app': app,
                'type': "http",
                'query_string': '',
                'headers': [
                    (b'host', b'foo.bar.baz')
                ],
                'fastapi_astack': stack,
                **scope
            }
        )
        result = runner.run_dependant(request, dependant, path=path)
        if inspect.isawaitable(result):
            result = await result

    return result



"""Declares :class:`JWKSEndpoint`."""
from ckms.types import JSONWebKeySet

import cbra
from cbra.cors import AnonymousReadCorsPolicy
from cbra.cors import CorsPolicyType


class JWKSEndpoint(cbra.Endpoint):
    """Single-tenant endpoint exposing the public keys used by the server to
    verify signatures and encrypt data, as a JSON Web Key Set (JWKS).
    """
    __module__: str = 'cbra.ext.jwks'
    cors_policy: type[CorsPolicyType] = AnonymousReadCorsPolicy
    name: str = 'metadata.jwks'
    description: str = (
        "Provides the JSON Web Key Set (JWKS) that a client may use to "
        "encrypt data (sent to the server) or verify signatures (issued "
        "by the server)."
    )
    method: str = 'GET'
    mount_path: str = '.well-known/jwks.json'
    summary: str = 'JSON Web Key Set (JWKS)'
    response_model: type[JSONWebKeySet] = JSONWebKeySet
    response_model_by_alias: bool = True
    response_description: str = "The server public keys."
    tags: list[str] = ["Server metadata"]
    with_options: bool = True

    async def handle(self) -> JSONWebKeySet:
        return self.request.jwks or JSONWebKeySet()



"""Declares :class:`OptionsMixin`."""
from typing import Any
from typing import Callable

import fastapi

import cbra
from cbra.types import ICorsPolicy
from cbra.types import IEndpoint
from cbra.cors import EndpointCorsPolicy
from .exceptions import Error


class Endpoint(cbra.Endpoint):
    cors_policy: type[ICorsPolicy] = EndpointCorsPolicy
    disable_options: bool = False
    methods: set[str]
    name: str
    state: str | None = None
    summary: str
    options_description: str

    @classmethod
    def add_to_router(
        cls,
        *,
        app: IEndpoint.RouterType,
        base_path: str = '/',
        method: str | None = None,
        cors_policy: type[ICorsPolicy] | None = None,
        request_handler: Callable[..., IEndpoint.ResponseTypes] | None = None,
        **kwargs: Any
    ) -> None:
        name = kwargs.pop('name', None) or cls.name
        cors_policy = cls.cors_policy.configure(
            allowed_methods=cls.methods | {"OPTIONS"},
            allowed_headers={
                "Content-Type"
            }
        )
        for method in cls.methods:
            super().add_to_router(
                app=app,
                base_path=base_path,
                cors_policy=cors_policy,
                method=method,
                request_handler=request_handler,
                name=name,
                **kwargs
            )

        if not cls.disable_options:
            Options = cbra.Endpoint.new(
                handle=cls.options
            )
            Options.add_to_router(
                app=app,
                base_path=base_path,
                cors_policy=cors_policy,
                request_handler=cls.options,
                method="OPTIONS",
                name=f'{name}.options',
                summary=cls.summary,
                description=cls.options_description,
                response_description="Allowed methods and CORS policy.",
                **kwargs
            )

    async def options(self) -> fastapi.Response:
        return fastapi.Response(
            headers={
                'Allow': str.join(',', sorted(self.cors_policy.allowed_methods))
             },
            status_code=200
        )

    @cbra.Endpoint.on_exception.register # type: ignore
    async def on_auth2_exception(
        self,
        exception: Error
    ) -> fastapi.Response:
        return await exception.as_response(
            request=self.request,
            state=self.state
        )

    @classmethod
    def exception_handler(cls, exception_class: type[Exception]):
        def decorator_factory(func: Callable[..., Any]):
            cbra.Endpoint.on_exception.register # type: ignore
            return func
        return decorator_factory


"""Declares :class:`EndpointCorsPolicy`."""
from typing import NoReturn

import fastapi

from cbra.utils import docstring
from cbra.types import IRequestHandler
from .basecorspolicy import BaseCorsPolicy


class EndpointCorsPolicy(BaseCorsPolicy):
    """A :class:`BaseCorsPolicy` implementation that defers the
    enforcement of the policy to the endpoint. This policy may
    be used when the endpoint has additional knowledge that is
    needed to determine if a cross-origin request is allowed to
    be served.

    The policy always accepts preflight requests that match the
    configured parameters.
    """
    __module__: str = 'cbra.cors'

    @docstring(BaseCorsPolicy)
    async def process_request(
        self,
        request: fastapi.Request,
        handler: IRequestHandler,
        origin: str | None = None
    ) -> bool | NoReturn:
        return await super().process_request(request, handler, origin)

    @docstring(BaseCorsPolicy)
    async def process_response(
        self,
        request: fastapi.Request,
        response: fastapi.Response,
        handler: IRequestHandler,
        origin: str | None = None
    ) -> None | NoReturn:
        if origin is None or origin == self.server:
            return

        if str.upper(request.method) == "OPTIONS":
            return await self.process_options_response(
                request=request,
                response=response,
                handler=handler
            )

        # It is assumed here that the endpoint rejected the request if
        # the CORS policy did not accept the origin or other parameters.
        # This basically means that CORS headers should be added allowing
        # all properties of the given response object.
        headers = {
            'Access-Control-Allow-Origin': self.origin,
            'Access-Control-Allow-Credentials': str(self.allow_credentials).lower()
        }
        response.headers.update(headers) # type: ignore

    async def process_options_response(
        self,
        request: fastapi.Request,
        response: fastapi.Response,
        handler: IRequestHandler,
        origin: str | None = None
    ) -> None | NoReturn:
        headers = {}
        headers['Access-Control-Allow-Credentials'] =\
            str(self.allow_credentials).lower()
        headers['Access-Control-Allow-Methods'] =\
            str.join(',', sorted(self.allowed_methods))
        headers['Access-Control-Allow-Origin'] = self.origin
        if self.allowed_headers:
            headers['Access-Control-Allow-Headers'] = str.join(
                ', ',
                sorted(self.allowed_headers)
            )
        response.headers.update(headers) # type: ignore


"""Declares :class:`Environment`."""
import os

from .dependency import Dependency


class Environment(Dependency):
    """A :class:`cbra.Dependency` implementation that points to an environment
    variable.
    """
    __module__: str = 'cbra.ext.ioc'

    def __init__(self, name: str):
        self.name = name
        super().__init__(use_cache=True)

    async def resolve(self):
        return os.getenv(self.name)



# Copyright (C) 2022 Cochise Ruhulessin <cochiseruhulessin@gmail.com>
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# No event shall the author(s) be liable for any claim or damages.
from .debugprincipal import DebugPrincipal
from .nullprincipal import NullPrincipal
from .subjectprincipal import SubjectPrincipal


__all__: list[str] = [
    'DebugPrincipal',
    'NullPrincipal',
    'SubjectPrincipal'
]


"""Declares :class:`JSONParser`."""
import json
import typing

import fastapi

from cbra.types import IParser
from ..exceptions import ParseError


class JSONParser(IParser):
    """
    Parses JSON-serialized data.
    """
    __module__: str = 'cbra.parsers'
    media_type: str = 'application/json'
    strict: bool    = True
    charset: str    = "utf-8"

    async def parse(self,
        request: fastapi.Request,
        media_type: typing.Optional[str] = None,
        parser_context: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> typing.Any:
        """Parses the incoming bytestream as JSON and returns the resulting
        data.
        """
        parser_context = parser_context or {}
        encoding = parser_context.get('encoding', self.charset)
        try:
            return json.loads(bytes.decode(await request.body(), encoding))
        except ValueError:
            raise ParseError



# pylint: skip-file
import json
from os import access
import typing
import urllib.parse

import httpx
import pytest
import yaml

import cbra
from cbra.types import IRenderer
from examples.endpoint import ExampleEndpoint


@pytest.fixture
def endpoint():
    return ExampleEndpoint


@pytest.fixture
def app(endpoint: typing.Type[ExampleEndpoint]):
    app = cbra.Application()
    app.add(endpoint, base_path='/get', method='GET')
    app.add(endpoint, base_path='/post', method='POST')
    return app


@pytest.mark.asyncio
async def test_endpoint_returns_parameter(
    client: httpx.AsyncClient
):
    response = await client.get(
        url='/get',
        headers={'Accept': "application/json"}
    )
    try:
        assert response.json().get('path_param') == None
    except json.JSONDecodeError:
        pytest.fail(response.text)


@pytest.mark.parametrize("renderer", ExampleEndpoint.renderers)
@pytest.mark.asyncio
async def test_content_negotiation_get(
    client: httpx.AsyncClient,
    renderer: typing.Type[IRenderer],
    url: str ='/get'
):
    response = await client.get(
        url=url,
        headers={'Accept': renderer.media_type}
    )
    assert response.status_code == 200
    assert 'Content-Type' in response.headers
    assert response.headers['Content-Type'] == renderer.response_media_type


@pytest.mark.parametrize("accept", [
    "application/*",
    "*/*"
])
@pytest.mark.asyncio
async def test_content_negotiation_wildcard(
    client: httpx.AsyncClient,
    accept: str,
    url: str ='/get'
):
    response = await client.get(
        url=url,
        headers={'Accept': accept}
    )
    assert response.status_code == 200
    assert 'Content-Type' in response.headers
    assert response.headers['Content-Type'] != accept


@pytest.mark.asyncio
async def test_unsupported_media_returns_406(
    client: httpx.AsyncClient,
    url: str = '/get'
):
    response = await client.get(
        url=url,
        headers={'Accept': "application/foo"}
    )
    assert response.headers['Content-Type'] != '*/*'
    assert response.status_code == 406, response.headers


@pytest.mark.asyncio
async def test_unsupported_media_used_default_renderer(
    client: httpx.AsyncClient,
    endpoint: typing.Type[ExampleEndpoint],
    url: str = '/get'
):
    response = await client.get(
        url=url,
        headers={'Accept': "application/foo"}
    )
    assert response.headers.get('Content-Type') == endpoint.fallback_renderer.media_type


@pytest.mark.parametrize("content_type,serialize", [ # type: ignore
    ("application/json", json.dumps),
    ("application/yaml", yaml.safe_dump), # type: ignore
    ("application/x-www-form-urlencoded", urllib.parse.urlencode),
])
@pytest.mark.asyncio
async def test_content_negotiation_with_body(
    client: httpx.AsyncClient,
    content_type: str,
    serialize: typing.Callable[..., str],
    url: str = '/post'
):
    dto = {'foo': 1}
    response = await client.post(
        url=url,
        content=str.encode(serialize(dto)),
        headers={
            'Accept': "application/json",
            'Content-Type': content_type,
        }
    )
    assert response.status_code == 200, response.text
    dto = response.json()
    assert dto.get('body') is not None
    assert int(dto['body'].get('foo')) == 1


@pytest.mark.asyncio
async def test_form_data_post(
    client: httpx.AsyncClient,
    url: str = '/post'
):
    dto = {'foo': 1}
    response = await client.post(
        url=url,
        data=dto,
        headers={'Accept': "application/json"}
    )
    assert response.status_code == 200, response.text
    dto = response.json()
    assert dto.get('body') is not None
    assert dto['body'].get('foo') == '1' # forms have no typing


@pytest.mark.parametrize("wants_digest,name", [
    ("sha-256", "sha-256"),
    ("sha-512", "sha-512"),
    ("sha-256,sha-512", "sha-256"),
    ("sha-256;q=0.5,sha-512;q=1", "sha-512"),
    (None, "sha-256"),
])
@pytest.mark.asyncio
async def test_digest_negotiation(
    client: httpx.AsyncClient,
    wants_digest: typing.Optional[str],
    name: str,
    url: str = '/get'
):
    headers = {}
    if wants_digest:
        headers['Wants-Digest'] = wants_digest
    response = await client.get(
        url,
        headers=headers
    )
    assert response.status_code == 200
    assert 'Digest' in response.headers
    assert str.startswith(response.headers['Digest'], name)


@pytest.mark.asyncio
async def test_digest_unsupported_is_rejected(
    client: httpx.AsyncClient,
    url: str = '/get'
):
    response = await client.get(
        url=url,
        headers={'Wants-Digest': "foo"}
    )
    assert response.status_code == 400


@pytest.mark.asyncio
async def test_digest_supported_is_selected(
    client: httpx.AsyncClient,
    url: str = '/get'
):
    response = await client.get(
        url=url,
        headers={'Wants-Digest': "foo,sha-256"}
    )
    assert response.status_code == 200, response.text


@pytest.mark.parametrize("wants_digest", [
    'foo;;,bar',
    'foo;q=a'
])
@pytest.mark.asyncio
async def test_digest_invalid_header_is_rejected(
    client: httpx.AsyncClient,
    wants_digest: str,
    url: str = '/get'
):
    response = await client.get(
        url=url,
        headers={'Wants-Digest': wants_digest}
    )
    dto = response.json()
    assert response.status_code == 400
    assert dto['spec']['code'] == 'INVALID_HEADER_VALUE'


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`Ping`."""
import pydantic


class Ping(pydantic.BaseModel):
    __module__: str = 'cbra.ext.woocommerce.models'

    webhook_id: int = pydantic.Field(
        default=...,
        title="Webhook ID",
        description=(
            "The `id` attribute of the webhook that is being pinged "
            "by WooCommerce."
        )
    )

    class Config:
        title: str = "Ping"


"""Declares :class:`Principal`."""
import dataclasses
import re
import typing

from ckms.types import ClaimSet
from ckms.types import MalformedPayload

from .types import IPrincipal


@dataclasses.dataclass
class Principal(IPrincipal):
    """A basic :class:`~cbra.ext.oauth2.IPrincipal` implementation."""
    sub: typing.Union[int, str]
    client_id: str
    scope: typing.Set[str] = dataclasses.field(
        default_factory=set
    )

    @classmethod
    def fromclaimset(cls, claims: ClaimSet) -> 'Principal':
        assert claims.sub is not None # nosec
        scope = []
        if 'client_id' not in claims.extra:
            raise MalformedPayload(
                message="The JWT must supply the 'client_id' claim."
            )
        if 'scope' in claims.extra:
            if not isinstance(claims.extra['scope'], str):
                raise MalformedPayload(
                    message="The 'scope' parameter must be a string."
                )
            scope = re.split('\\s', claims.extra['scope'])
        return cls(
            client_id=claims.extra['client_id'],
            sub=claims.sub,
            scope=set(scope)
        )

    def has_scope(
        self,
        scope: typing.Union[str, typing.Set[str]]
    ) -> bool:
        """Return a boolean indicating if the :class:`IPrincipal` is
        authorized to use the given `scope`.
        """
        if isinstance(scope, str):
            scope = {scope}
        return self.scope >= scope


"""Declares :class:`AuthorizationServer`."""
import enum
import pathlib
import types
import typing

import ckms
import ckms.jose
import fastapi
from ckms.core import Keychain
from ckms.types import IKeychain
from ckms.types import JSONWebKeySet
from fastapi.exceptions import RequestValidationError
from unimatrix.exceptions import CanonicalException

import cbra
from cbra.auth import NullPrincipal
from cbra.ext import ioc
from cbra.types import IEndpoint
from cbra.types import IRouteable
from .apiroute import APIRoute
from .authorizeendpoint import AuthorizationEndpoint
from .exceptions import Error
from .jwksendpoint import JWKSEndpoint
from .memorystorage import MemoryStorage
from .metadataendpoint import MetadataEndpoint
from .nullclientrepository import NullClientRepository
from .nullsubjectrepository import NullSubjectRepository
from .pushedauthorizationrequestendpoint import PushedAuthorizationRequestEndpoint
from .types import IAuthorizeEndpoint
from .types import IStorage
from .types import IClientRepository
from .types import IOpenAuthorizationServer
from .types import IPrincipal
from .types import ISubjectRepository
from .types import ITokenIssuer
from .types import IUpstreamProvider
from .types import IUpstreamReturnHandler
from .types import ServerMetadata
from .tokenissuer import TokenIssuer
from .tokenrequesthandler import TokenRequestHandler


DEFAULT_TAGS: list[enum.Enum | str] = ['OAuth 2.1/OpenID Connect 1.0']
SUMMARY = pathlib.Path(__file__).parent.joinpath('authorizationserver.md')


class AuthorizationServer(fastapi.APIRouter, IOpenAuthorizationServer):
    __module__: str = 'cbra.ext.oauth2'
    _jwks: JSONWebKeySet | None = None
    allowed_signing_algorithms: list[str]
    allowed_content_encryption: list[str]
    asgi: cbra.Application
    authorize: type[IAuthorizeEndpoint]
    authorize_endpoint_name: str = 'oauth2.authorize'
    callbacks: dict[str, type[IUpstreamReturnHandler]] = {}
    clients: type[IClientRepository]
    decryption_alg: str
    decryption_key: str
    enable_authorization: bool
    enable_par: bool
    issuer: type[ITokenIssuer]
    jwks_endpoint_name: str
    jwks_tags: list[str] = ['oauth2']
    login_endpoint: str | None
    login_url: str | None
    metadata: ServerMetadata = ServerMetadata(
        token_endpoint_auth_methods_supported=['private_key_jwt']
    )
    metadata_endpoint_name: str = 'oauth2.metadata'
    metadata_overrides: dict[str, bool | str | list[str]]
    par: type[PushedAuthorizationRequestEndpoint] = PushedAuthorizationRequestEndpoint
    principal_factory: typing.Callable[..., IPrincipal] | type[object]
    providers: dict[str, IUpstreamProvider] = {}
    require_par: bool
    server_name: str | None = None
    signing_alg: str
    signing_key: str
    storage: type[IStorage]
    summary: str = open(SUMMARY).read()
    subjects: type[ISubjectRepository]
    token: type[TokenRequestHandler]

    @property
    def codec(self) -> ckms.jose.PayloadCodec:
        """The :class:`ckms.jose.PayloadCodec` instance that is used by
        the authorization server to sign JWS objects and decrypt JWE
        objects.
        """
        return ckms.jose.PayloadCodec(
            decrypter=self.keychain,
            signer=self.keychain
        )

    @property
    def jwks(self) -> JSONWebKeySet:
        """The JSON Web Key Set (JWKS) containing the public keys used
        by the authorization server."""
        if self._jwks is None:
            self._jwks = self.keychain.tagged(self.jwks_tags).as_jwks(private=False)
        return self._jwks

    @property
    def keychain(self) -> IKeychain:
        """The keychain used by the authorization server to
        sign and decrypt.
        """
        return self.asgi.keychain

    def add_to_router( # type: ignore
        self,
        *,
        app: IRouteable.RouterType,
        base_path: str,
        **kwargs: str
    ) -> None:
        # Configure the OAuth 2.0 server metadata and OpenID Connect
        # Discovery endpoints.
        app = typing.cast(cbra.Application, app)
        if self.server_name is not None:
            raise NotImplementedError
        self.metadata_endpoint.add_to_router(
            app=app,
            base_path='/.well-known/oauth-authorization-server',
            cors_policy=cbra.cors.AnonymousReadCorsPolicy,
            name='oauth2.metadata',
            tags=DEFAULT_TAGS,
            response_model_exclude_defaults=True
        )
        if self.authorize is not None and self.enable_authorization:
            self.authorize.add_to_router( # type: ignore
                app=self,
                base_path='/authorize',
                name='oauth2.authorize'
            )
            self.par.add_to_router(
                app=self,
                base_path='/par',
                name='oauth2.par'
            )

        # Add the JWKS endpoint for this specific API route.
        JWKSEndpoint.add_to_router(
            app=self,
            base_path='/jwks.json',
            name='oauth2.jwks',
            response_model_exclude_defaults=True
        )

        if self.token is not None:
            self.token.add_to_router(
                app=self,
                base_path='/token'
            )

        for cb in self.callbacks.values():
            cb.add_to_router(
                app=self,
                base_path='/callback'
            )

        self.asgi = app
        app.include_router(
            router=self,
            prefix=base_path,
            tags=DEFAULT_TAGS
        )
        app.add_exception_handler(Error, self.on_oauth2_exception) # type: ignore
        app.add_event_handler('startup', self.setup_authorization_server) # type: ignore

        assert app.openapi_tags is not None # nosec
        app.openapi_tags.append({
            'name': DEFAULT_TAGS[0],
            'description': self.summary
        })

    @classmethod
    def fromsettings(cls, settings: types.ModuleType) -> 'AuthorizationServer':
        """Create a new :class:`AuthorizationServer` from a settings module.

        The following members may be defined to configure the server:

        - `OAUTH_SERVER_CONFIG` - A dictionary holding configuration parameters
          for the authorization server e.g.:

          .. code:: python

            OAUTH_SERVER_CONFIG = {
                'signing_key': 'sig
            }

        - `OAUTH_CLIENTS` - Either a string holding the qualified name of a
          class that implements the :class:`cbra.ext.oauth2.types.IClientRepository`
          interface, or a dictionary specifying at least a `class` attribute
          holding the qualified name of the implementation, and additional
          members specifying configuration parameters.
        - `OAUTH_CLIENT_MODEL` - A string holding the qualified name to a Python
          class that is used as the :class:`cbra.ext.oauth2.types.IClient`
          implementation returned from the clients repository.
        """
        raise NotImplementedError

    def __init__(
        self,
        *,
        signing_key: str,
        allowed_content_encryption: list[str] | None = None,
        allowed_signing_algorithms: list[str] | None = None,
        authorize: type[IAuthorizeEndpoint | IEndpoint] = AuthorizationEndpoint,
        callbacks: list[type[IUpstreamReturnHandler]] | None = None,
        clients: type[IClientRepository] | None = None,
        grant_types: set[str] | None = None,
        jwks_tags: list[str] = ['oauth2'],
        jwks_endpoint_name: str = 'metadata.jwks',
        login_endpoint: str | None = None,
        login_url: str | None = None,
        metadata_overrides: dict[str, bool | str | list[str]] | None = None,
        storage: type[IStorage] | None = None,
        subjects: type[ISubjectRepository] = NullSubjectRepository,
        principal_factory: typing.Callable[..., IPrincipal] | type[object] = NullPrincipal,
        providers: list[IUpstreamProvider] | None = None,
        server_name: str | None = None,
        issuer: type[ITokenIssuer] = TokenIssuer,
        token: type[TokenRequestHandler] = TokenRequestHandler,
        par: type[PushedAuthorizationRequestEndpoint] = PushedAuthorizationRequestEndpoint,
        enable_authorization: bool = True,
        enable_par: bool = True,
        require_par: bool = False,
        **kwargs: typing.Any
    ):
        kwargs.setdefault('route_class', APIRoute.with_server(self))
        super().__init__(**kwargs)

        # If the authorization_code grant is not allowed, then there is no point
        # in enabling the authorization endpoints.
        if "authorization_code" not in (grant_types or []):
            enable_authorization = False

        self.allowed_content_encryption = allowed_content_encryption or ['A256GCM']
        self.allowed_signing_algorithms = allowed_signing_algorithms or ['EdDSA', 'PS256']
        self.authorize = typing.cast(
            type[IAuthorizeEndpoint],
            authorize.new(principal_factory=principal_factory) # type: ignore
        )
        self.callbacks = {cb.provider: cb for cb in (callbacks or [])}
        self.clients = clients or NullClientRepository
        self.enable_authorization = enable_authorization
        self.enable_par = enable_authorization and (enable_par or require_par)
        self.issuer = issuer
        self.jwks_endpoint_name = jwks_endpoint_name
        self.jwks_tags = jwks_tags
        self.login_endpoint = login_endpoint
        self.login_url = login_url
        self.metadata_endpoint = MetadataEndpoint.new(server=self)
        self.metadata_endpoint_name = 'oauth2.metadata'
        self.metadata_overrides = metadata_overrides or {}
        self.par = par
        self.providers = {
            provider.name: provider 
            for provider in (providers or [])
        }
        self.require_par = self.enable_par and require_par
        self.server_name = server_name
        self.signing_key = signing_key
        self.storage = storage or MemoryStorage
        self.subjects = subjects
        self.token = token

        ioc.provide('cbra.ext.oauth2.ClientRepository', self.clients, True)
        ioc.provide('cbra.ext.oauth2.SubjectRepository', self.subjects, True)
        ioc.provide('cbra.ext.oauth2.TransientStorage', self.storage, True)
        self.metadata.grant_types_supported = list(sorted(grant_types or []))

    def allows_grant(self, grant_type: str) -> bool:
        """Return a boolean indicating if the server is configured to allow
        the given grant type.
        """
        return grant_type in (self.metadata.grant_types_supported or [])

    def get_jwks(self) -> JSONWebKeySet:
        """Return the :class:`ckms.jose.models.JSONWebKeySet` that is used
        by the server.
        """
        assert self.jwks is not None # nosec
        return self.jwks

    def reverse(self, request: fastapi.Request, name: str) -> str:
        return request.url_for(name)

    async def get_metadata(
        self,
        request: fastapi.Request
    ) -> ServerMetadata:
        assert self.jwks is not None # nosec
        metadata = {
            **self.metadata.dict(),
            'issuer': f'{request.url.scheme}://{request.url.netloc}',
            'token_endpoint': request.url_for('oauth2.token') if self.token else None,
            'jwks_uri': self.reverse(request, 'oauth2.jwks'),
            'require_pushed_authorization_requests': self.require_par
        }
        if self.enable_authorization:
            metadata['authorization_endpoint'] = request.url_for('oauth2.authorize')
        if self.enable_par:
            metadata['request_parameter_supported'] = True
            metadata['request_uri_parameter_supported'] = True
            metadata['request_object_encryption_alg_values_supported'] = self.jwks.get_encryption_algorithms()
            metadata['request_object_encryption_enc_values_supported'] = self.allowed_content_encryption
            metadata['request_object_signing_alg_values_supported'] = self.allowed_signing_algorithms
            metadata['require_request_uri_registration'] = True
            metadata['pushed_authorization_request_endpoint'] = request.url_for('oauth2.par')
        if self.token:
            metadata['token_endpoint_auth_signing_alg_values_supported'] = self.allowed_signing_algorithms
        metadata.update(self.metadata_overrides)
        return ServerMetadata(**metadata)

    async def add_encryption_key(self, name: str, params: dict[str, typing.Any]) -> None:
        """Add an encryption key that is specifically used by the authorization
        server.
        """
        tags = params.setdefault('tags', [])
        for tag in self.jwks_tags:
            if tag in tags:
                continue
            tags.append(tag)
        return await self.asgi.add_encryption_key(name, params)

    async def add_signing_key(self, name: str, params: dict[str, typing.Any]) -> None:
        """Add a signing key that is specifically used by the authorization
        server.
        """
        tags = params.setdefault('tags', [])
        for tag in self.jwks_tags:
            if tag in tags:
                continue
            tags.append(tag)
        return await self.asgi.add_signing_key(name, params)

    async def metadata_redirect(self, request: fastapi.Request):
        return fastapi.responses.RedirectResponse(
            self.reverse(request, self.metadata_endpoint_name),
            status_code=303
        )

    async def setup_authorization_server(self):
        await self.keychain

    async def on_exception(
        self,
        request: fastapi.Request,
        exception: CanonicalException | RequestValidationError
    ) -> fastapi.Response:
        raise exception

    async def on_oauth2_exception(
        self,
        request: fastapi.Request,
        exception: Error
    ) -> fastapi.Response:
        return await exception.as_response(request)


# pylint: skip-file
import typing

import pydantic

import cbra


class FooModel(pydantic.BaseModel):
    foo: int


class BarModel(pydantic.BaseModel):
    bar: int


BodyModel = typing.Union[
    FooModel,
    BarModel
]


class ContentNegotiationEndpoint(cbra.Endpoint):
    model = BodyModel

    async def handle(self, dto: BodyModel):
        return dto.dict()


"""Declares :class:`AuthorizationCode`."""


class AuthorizationCode:
    __module__: str = 'cbra.ext.oauth2.types'
    code: str
    request_id: str

    def __init__(self, request_id: str, code: str):
        self.request_id = request_id
        self.code = code


# pylint: skip-file
import pytest_asyncio
from ckms.core import Keychain
from httpx import AsyncClient

import cbra
from cbra.ext.jwks import JWKSEndpoint
from cbra.ext.oauth2 import AuthorizationServer


base_url: str = "https://cbra-oauth2.localhost"


@pytest_asyncio.fixture # type: ignore
async def app(
    server: AuthorizationServer,
    keychain: Keychain
) -> cbra.Application:
    await keychain
    app = cbra.Application(
        keychain=keychain
    )
    app.add(JWKSEndpoint)
    app.add(server, base_path='/oauth2')
    await app.on_startup()
    return app


@pytest_asyncio.fixture # type: ignore
async def client(app: cbra.Application):
    params = {
        'app': app,
        'base_url': base_url
    }
    async with AsyncClient(**params) as client:
        yield client


"""Declares :class:`ConfigFileClientRepository`."""
import os
import pathlib
import re
import typing

import aiofiles
import yaml

from .clientconfig import ClientConfig
from .types import IClient
from .types import IClientRepository


class ConfigFileClientRepository(IClientRepository):
    """A :class:`~cbra.ext.oauth2.types.IClientRepository` implementation
    that loads client configuration from a directory on the local fileststem.
    """
    __module__: str = 'cbra.ext.oauth2'
    base_dir: pathlib.Path = pathlib.Path(
        os.path.join(os.getcwd(), 'etc/oauth2')
    )
    caching: bool = True
    clients: dict[str, IClient] = {}
    model: type[IClient] = ClientConfig
    pattern: re.Pattern[str] = re.compile(r'^[A-Za-z0-9\-\_]+$')

    def __init__(self) -> None:
        self.clients = {}
        if self.caching:
            self.clients = ConfigFileClientRepository.clients

    async def exists(self, client_id: str) -> bool:
        """Retur a boolean indicating if the client exists."""
        if not self.pattern.match(client_id):
            return False
        fn = self.base_dir.joinpath(f'{client_id}.yml')
        return fn.exists()

    async def get(self, client_id: str) -> IClient:
        """Lookup an OAuth 2.0 client from a configuration file
        stored in the :attr:`ConfigFileClientRepository.base_dir`
        directory.
        """
        if not self.pattern.match(client_id):
            raise self.ClientDoesNotExist
        if client_id not in self.clients:
            fn = self.base_dir.joinpath(f'{client_id}.yml')
            if not fn.exists():
                raise self.ClientDoesNotExist
            async with aiofiles.open(fn, 'r') as f: # type: ignore
                params = yaml.safe_load(await f.read()) # type: ignore
                params = typing.cast(dict[str, typing.Any], params)
            client = self.model.fromdict({**params, 'client_id': client_id})
            self.clients[client_id] = client
        return self.clients[client_id]


"""Declares :class:`TokenRequest`."""
import typing

import pydantic

from .authorizationcodegrant import AuthorizationCodeGrant
from .basegrant import BaseGrant
from .clientcredentialsgrant import ClientCredentialsGrant
from .jwtbearerassertiongrant import JWTBearerAssertionGrant


class TokenRequestParameters(pydantic.BaseModel):
    """The parameters available for requests to the **Token Endpoint**."""
    __module__: str = 'cbra.ext.oauth2'
    __root__: typing.Union[
        AuthorizationCodeGrant,
        ClientCredentialsGrant,
        JWTBearerAssertionGrant,
    ]

    def get_root(self) -> BaseGrant:
        return self.__root__


"""Declares :class:`AuthorizationCodeGrant`."""
import typing

import pydantic

from .basegrant import BaseGrant
from .granttype import GrantType


class AuthorizationCodeGrant(BaseGrant):
    __module__: str = 'cbra.ext.oauth2.types'

    grant_type: typing.Literal[GrantType.authorization_code] = pydantic.Field(
        default=...,
        title="Grant type",
        description=(
            "Must be `authorization_code`."
        ),
        example="authorization_code"
    )

    code: str = pydantic.Field(
        default=...,
        title="Authorization code",
        description=(
            "The authorization code that was received from the authorization "
            "server through user-agent redirection."
        )
    )

    redirect_uri: str = pydantic.Field(
        default=None,
        title="Redirect URI",
        description=(
            "Required if the `redirect_uri` parameter was included in the "
            "authorization request and their values must be identical, else "
            "this parameter must be absent from the request."
        )
    )


"""Declares :class:`MessagePublisher`."""
from cbra.ext.ioc import Dependency
from .transaction import Transaction


class MessagePublisher(Dependency):
    """A :class:`cbra.Dependency` implementation that provides a transactional
    message publisher.
    """
    __module__: str = 'cbra'

    def __init__(self):
        Dependency.__init__(self, use_cache=True)

    async def resolve(self): # pragma: no cover
        async with Transaction() as tx:
            yield tx


_default: MessagePublisher = MessagePublisher()



"""Declares :class:`FileDependencySpec`."""
import typing

import aiofiles

from .dependencyspec import DependencySpec


class FileDependencySpec(DependencySpec):
    """The configuration of a dependency that is loaded from a file."""
    __module__: str = 'cbra.ext.ioc.models'
    type: typing.Literal['file'] = 'file'
    path: str

    async def resolve(self) -> typing.Any:
        """Resolve the dependency specified by the input parameters."""
        async with aiofiles.open(self.path, 'rb') as fp: # pylint: disable=C0103
            return await fp.read()



# pylint: skip-file
from .basequerymodel import BaseQueryModel
from .iapplication import IApplication
from .icanonicalexception import ICanonicalException
from .icontentnegotiation import IContentNegotiation
from .icorspolicy import ICorsPolicy
from .iendpoint import IEndpoint
from .iparser import IParser
from .iprincipal import IPrincipal
from .iqueryrunner import IQueryRunner
from .irenderer import IRenderer
from .irequesthandler import IRequestHandler
from .iresponsedigest import IResponseDigest
from .irouteable import IRouteable
from .iuseragent import IUserAgent
from .jsonqueryparameter import JSONQueryParameter
from .nullquerymodel import NullQueryModel
from .queryoptions import QueryOptions
from .request import Request


__all__ = [
    'BaseQueryModel',
    'IApplication',
    'ICanonicalException',
    'IContentNegotiation',
    'ICorsPolicy',
    'IEndpoint',
    'IParser',
    'IPrincipal',
    'IQueryRunner',
    'IRenderer',
    'IRequestHandler',
    'IResponseDigest',
    'IRouteable',
    'IUserAgent',
    'JSONQueryParameter',
    'NullQueryModel',
    'QueryOptions',
    'Request',
]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebKeySet


class Request(fastapi.Request):

    @property
    def codec(self) -> PayloadCodec:
        """The :class:`~ckms.jose.PayloadCodec` instance used to
        decode JOSE objects.
        """
        return self.app.codec

    @property
    def keychain(self) -> Keychain:
        """The :class:`~ckms.core.Keychain` instance holding the
        signing and decryption keys used by the application.
        """
        return self.app.keychain

    @property
    def jwks(self) -> JSONWebKeySet:
        """A :class:`~ckms.types.JSONWebKeySet` holding the public keys intended
        for external consumers.
        """
        return self.keychain.tagged('unimatrixone.io/public-keys').as_jwks()


# pylint: skip-file
from .authorizationcode import AuthorizationCode
from .authorizationcodegrant import AuthorizationCodeGrant
from .authorizationrequest import AuthorizationRequest
from .authorizationrequestparameters import AuthorizationRequestParameters
from .baseauthorizationrequest import BaseAuthorizationRequest
from .basegrant import BaseGrant
from .basesubject import BaseSubject
from .clientassertiontype import ClientAssertionType
from .clientcredentialsgrant import ClientCredentialsGrant
from .configurable import Configurable
from .dependantmodel import DependantModel
from .fields import GrantTypeField
from .fields import ResourceField
from .granttype import GrantType
from .iauthorizeendpoint import IAuthorizeEndpoint
from .iauthorizationserver import IAuthorizationServer
from .iclient import IClient
from .iclientrepository import IClientRepository
from .iopenauthorizationserver import IOpenAuthorizationServer
from .iprincipal import IPrincipal
from .istorage import IStorage
from .isubject import ISubject
from .isubjectrepository import ISubjectRepository
from .itokenissuer import ITokenIssuer
from .iupstreamprovider import IUpstreamProvider
from .iupstreamreturnhandler import IUpstreamReturnHandler
from .jar import JAR
from .jwtbearerassertiongrant import JWTBearerAssertionGrant
from .redirecturl import RedirectURL
from .responsetype import ResponseType
from .rfc9068token import RFC9068Token
from .servermetadata import ServerMetadata
from .spaceseparatedlist import SpaceSeparatedList
from .stringorlist import StringOrList
from .tokenrequestparameters import TokenRequestParameters
from .tokenresponse import TokenResponse
from .tokentype import TokenType


__all__: list[str] = [
    'AuthorizationCode',
    'AuthorizationCodeGrant',
    'AuthorizationRequest',
    'AuthorizationRequestParameters',
    'BaseAuthorizationRequest',
    'BaseGrant',
    'BaseSubject',
    'ClientAssertionType',
    'ClientCredentialsGrant',
    'Configurable',
    'DependantModel',
    'GrantType',
    'GrantTypeField',
    'IAuthorizeEndpoint',
    'IAuthorizationServer',
    'IClient',
    'IClientRepository',
    'IOpenAuthorizationServer',
    'IPrincipal',
    'IStorage',
    'ISubject',
    'ISubjectRepository',
    'ITokenIssuer',
    'IUpstreamProvider',
    'IUpstreamReturnHandler',
    'JAR',
    'JWTBearerAssertionGrant',
    'RedirectURL',
    'ResourceField',
    'ResponseType',
    'RFC9068Token',
    'ServerMetadata',
    'SpaceSeparatedList',
    'StringOrList',
    'TokenRequestParameters',
    'TokenResponse',
    'TokenType',
]


# pylint: skip-file
import typing

import pytest
import pytest_asyncio
from ckms.core import Keychain
from httpx import AsyncClient

import cbra
from cbra.ext.oauth2 import AuthorizationServer
from cbra.ext.oauth2.types import ISubject


base_url: str = "https://cbra-oauth2.localhost"


@pytest_asyncio.fixture # type: ignore
async def app(
    server: AuthorizationServer,
    keychain: Keychain
) -> cbra.Application:
    await keychain
    app = cbra.Application(
        keychain=keychain
    )
    app.add(server, base_path='/oauth2') # type: ignore
    await app.on_startup()
    return app


@pytest.fixture
def client_kwargs(app: cbra.Application, bob: ISubject) -> dict[str, typing.Any]:
    return {
        'app': app,
        'base_url': base_url,
        'cookies': {
            'subject': bob.sub
        }
    }


@pytest_asyncio.fixture # type: ignore
async def client(client_kwargs: dict[str, typing.Any]):
    async with AsyncClient(**client_kwargs) as client:
        yield client